/*
    FFT Display - browser version 1.0
    7 July, 2015 - updated 2 April, 2019
    Copyright © 2010-2019 Harry Whitfield
    mailto:g6auc@arrl.net
*/

/*jslint browser, eval, this */

/*property
    PI, formula, height, isNaN, length, modulus, multiplier, now, prototype, re,
    replace, scale, title, toFixed, toShortest, units, value, width
*/

/*global fft, report, command, makeBackground, Complex,
	multiplierPref, scalePref, unitsPref, barChart, plotBar, show,
	alert
*/

var graphBgColor;
var graphColor;
var axisColor;
var axisTextColor;
var reportBgColor;
var reportTextColor;

Number.prototype.toShortest = function (precision) {
    "use strict";
    return this.toFixed(precision).replace(/\.0+$/, "").replace(/(\d+\.\d*[1-9])0*$/, "$1");
};

function updatePrefs() {
    "use strict";
    graphBgColor = "#f5f8b8";
    graphColor = "#000000";
    axisColor = "#000000";
    axisTextColor = "#000000";

    reportBgColor = "#84e67c";
    reportTextColor = "#000000";
}
updatePrefs();

function rescale() {
    "use strict";
    alert("rescale");
}

function openBarChart(canvas) {
    "use strict";
    var width = canvas.width;
    var height = canvas.height;

    makeBackground(canvas, width, height, graphBgColor, axisColor);
}

var waveTitle = [
    "sine wave",
    "two sine waves",
    "modulated sine wave",
    "amplitude modulated sine wave",
    "frequency modulated sine wave",
    "sine wave exponentially damped",
    "square wave",
    "sine wave pulse",
    "sine of t-squared",
    "sine wave log multiplied",
    "sinc function",
    "sine of an exponential",
    "sine squared",
    "multiple pulse",
    "square pulse",
    "sawtooth",
    "triangle",
    "gaussian"
];

var waveFormula = [
    "sin(2*pi*128*t)",
    "sin(2*pi*100*t) + 2*sin(2*pi*150*t)",
    "sin(2*pi*140*t)*sin(2*pi*20*t)",
    "sin(2*pi*100*t)*(1 - cos(2*pi*5*t))",
    "sin(2*pi*100*t + 2.41*cos(2*pi*2*t))",
    "sin(2*pi*128*t)*exp(-2*2*pi*t)",
    "(sin(2*pi*32*t)>0)?1:-1",
    "(2*pi*t<0.2)?sin(2*pi*128*t):0",
    "sin(4*2*pi*t*2*pi*t)",
    "sin(2*pi*100*t)*log(2*pi*t)",
    "sin(2*pi*100*t)/(2*pi*100*t)",
    "sin(exp(2*pi*t))",
    "sin(2*pi*t)*sin(2*pi*t)-0.5",
    "sin(200*t)*(exp(-20*(t-1)*(t-1))+exp(-20*(t-3)*(t-3))+exp(-20*(t-5)*(t-5)))",
    "(t<0.1)?1:0",
    "2*t%1-0.5",
    "(8*t%2<1)?8*t%1-0.5:0.5-(8*t%1)",
    "exp(-(t-4)*(t-4)*1000)"
];

function write(data) {
    "use strict";
    report.value += data;
}

function writeln(data) {
    "use strict";
    report.value += data + "\n";
}

function makeFFT(formula, multiplier, scale, units, title) {
    "use strict";
    var data;
	var max;
	var maxf;
	var u;
	var val;
	var xBase;
	var yBase;
	var t1;
	var t2;
	var font;
	var tmp;

    function makeData(formula, multiplier, scale) {
        var k;
		var theData = [];
		var pi = Math.PI;
		var t;

        k = 0;
        while (k < 1024) {
            t = scale * k / (1024 * multiplier);

            theData[k] = new Complex(eval(formula), 0); // formula is a function of t, pi and Math functions.
            if (Number.isNaN(theData[k].re)) {
                theData[k].re = 0;
            }

            k += 1;
        }
        return theData;
    }

    function siPrefix(m) {
        switch (m) {
        case 1e0:
            return "";
        case 1e3:
            return "k";
        case 1e6:
            return "M";
        case 1e9:
            return "G";
        case 1e12:
            return "T";
        default:
            return "?";
        }
    }

    function mathReplace(formula) {
        return formula.replace(/abs\(/g, "Math.abs(").replace(/(acos|cos)\(/g, "Math.$1(").replace(/(asin|sin)\(/g, "Math.$1(").replace(/(atan2|atan|tan)\(/g, "Math.$1(").replace(/ceil\(/g, "Math.ceil(").replace(/exp\(/g, "Math.exp(").replace(/floor\(/g, "Math.floor(").replace(/log\(/g, "Math.log(").replace(/max\(/g, "Math.max(").replace(/min\(/g, "Math.min(").replace(/pow\(/g, "Math.pow(").replace(/random\(/g, "Math.random(").replace(/round\(/g, "Math.round(").replace(/sqrt\(/g, "Math.sqrt(");
    }

    if (multiplier === undefined) {
        multiplier = multiplierPref;
    }
    if (scale === undefined) {
        scale = scalePref;
    }
    if (units === undefined) {
        units = unitsPref;
    }
    if (title === undefined) {
        title = "Untitled formula";
    }

    data = makeData(mathReplace(formula), multiplier, scale);

    units = siPrefix(multiplier) + units;

    writeln("--- " + title + " ---");
    writeln("Formula:\t" + formula);
    writeln("Multiplier:\t" + multiplier);
    writeln("Scale:\t\t" + scale.toShortest(3));
    writeln("Units:\t\t" + units);

    t1 = Date.now();
    fft(data, 10, false);
    t2 = Date.now();
    writeln("Calculation:\t" + (t2 - t1) + "ms");

    max = 0;
    u = 0;
    while (u < 512) {
        val = data[u].modulus() / 512;
        if (val > max) {
            max = val;
            maxf = u;
        }
        u += 1;
    }

    tmp = maxf / scale;
    writeln("Maximum:\t" + max.toFixed(3) + " at freq: " + tmp.toShortest(3) + units);

    openBarChart(barChart);
    xBase = 28;
    yBase = 28;
    plotBar(barChart, xBase, yBase + 200, 512, 1, axisColor);

    font = "12px Courier";

    u = 0;
    while (u < 512) {
        val = data[u].modulus() / 512;
        plotBar(barChart, xBase + u, yBase + 200, 1, -200 * val / max, graphColor);

        if ((u % 10) === 0) {
            if ((u % 100) === 0) {
                plotBar(barChart, xBase + u, yBase + 200, 1, 10, axisColor);
                if (u !== 0) {
                    tmp = u / scale;
                    show(barChart, tmp.toShortest(3), xBase + u - 9, yBase + 220, font);
                } else {
                    tmp = u / scale;
                    show(barChart, tmp.toShortest(3), xBase + u - 3, yBase + 220, font);
                }
            } else if ((u % 50) === 0) {
                plotBar(barChart, xBase + u, yBase + 200, 1, 6, axisColor);
            } else {
                plotBar(barChart, xBase + u, yBase + 200, 1, 3, axisColor);
            }
        }
//      report.value += (u + ": " + val.toFixed(3) + (val === max ? "  **" : ""));
        u += 1;
    }
    show(barChart, units, xBase + 240, yBase + 220, font);
}

function calcFFT(params) {
    "use strict";
    var formula;
	var multiplier;
	var scale;
	var units;
	var title;

    function value(m) {
        switch (m) {
        case "unity":
            return 1e0;
        case "k":
            return 1e3;
        case "M":
            return 1e6;
        case "G":
            return 1e9;
        case "T":
            return 1e12;
        }
    }

    if (params) {
        formula = params.formula;
        multiplier = value(params.multiplier);
        scale = Number(params.scale);
        units = params.units;
        title = params.title;

        if (formula !== "") {
            command.value = "makeFFT(\"" + formula + "\", " + multiplier + ", " + scale.toFixed(1) + ", \"" + units + "\", \"" + title + "\");";
            makeFFT(formula, multiplier, scale, units, title);
        }
    }
}

function testFFT(n) {
    "use strict";
    var multiplier;
	var scale;
	var units;

    if ((n < 0) || (n >= waveFormula.length)) {
        n = 0;
    }
    multiplier = 1.0;
    scale = scalePref;
    if (n < 13) {
        units = "Hz";
    } else {
        units = "rad/s";
    }
    command.value = "makeFFT(\"" + waveFormula[n] + "\", " + multiplier + ", " + scale.toFixed(1) + ", \"" + units + "\", \"" + waveTitle[n] + "\");";
    makeFFT(waveFormula[n], multiplier, scale, units, waveTitle[n]);
}
